import { DocNode, DocNodeKind, IDocNodeParameters } from './DocNode';
import { DocMemberReference } from './DocMemberReference';
import { Excerpt } from '../parser/Excerpt';
/**
 * Constructor parameters for {@link DocDeclarationReference}.
 */
export interface IDocDeclarationReferenceParameters extends IDocNodeParameters {
    packageNameExcerpt?: Excerpt;
    packageName?: string;
    importPathExcerpt?: Excerpt;
    importPath?: string;
    importHashExcerpt?: Excerpt;
    memberReferences?: DocMemberReference[];
}
/**
 * Represents a declaration reference.
 *
 * @remarks
 * Declaration references are TSDoc expressions used by tags such as `{@link}`
 * or `{@inheritDoc}` that need to refer to another declaration.
 */
export declare class DocDeclarationReference extends DocNode {
    /** {@inheritDoc} */
    readonly kind: DocNodeKind;
    private _packageNameParticle;
    private _importPathParticle;
    private _importHashParticle;
    private _memberReferences;
    /**
     * Don't call this directly.  Instead use {@link TSDocParser}
     * @internal
     */
    constructor(parameters: IDocDeclarationReferenceParameters);
    /**
     * The optional package name, which may optionally include an NPM scope.
     *
     * Example: `"@scope/my-package"`
     */
    readonly packageName: string | undefined;
    /**
     * The optional import path.  If a package name is provided, then if an import path is provided,
     * the path must start with a "/" delimiter; otherwise paths are resolved relative to the source file
     * containing the reference.
     *
     * Example: `"/path1/path2"`
     * Example: `"./path1/path2"`
     * Example: `"../path2/path2"`
     */
    readonly importPath: string | undefined;
    /**
     * The chain of member references that indicate the declaration being referenced.
     * If this list is empty, then either the packageName or importPath must be provided,
     * because the reference refers to a module.
     */
    readonly memberReferences: ReadonlyArray<DocMemberReference>;
    /** @override */
    updateParameters(parameters: IDocDeclarationReferenceParameters): void;
    /**
     * {@inheritDoc}
     * @override
     */
    getChildNodes(): ReadonlyArray<DocNode>;
}
