import { DocNode, DocNodeKind, IDocNodeParameters } from './DocNode';
import { Excerpt } from '../parser/Excerpt';
import { DocMemberIdentifier } from './DocMemberIdentifier';
import { DocMemberSymbol } from './DocMemberSymbol';
import { DocMemberSelector } from './DocMemberSelector';
/**
 * Constructor parameters for {@link DocMemberReference}.
 */
export interface IDocMemberReferenceParameters extends IDocNodeParameters {
    hasDot: boolean;
    dotExcerpt?: Excerpt;
    leftParenthesisExcerpt?: Excerpt;
    memberIdentifier?: DocMemberIdentifier;
    memberSymbol?: DocMemberSymbol;
    colonExcerpt?: Excerpt;
    selector?: DocMemberSelector;
    rightParenthesisExcerpt?: Excerpt;
}
/**
 * A {@link DocDeclarationReference | declaration reference} includes a chain of
 * member references represented using `DocMemberReference` nodes.
 *
 * @remarks
 * For example, `example-library#ui.controls.Button.(render:static)` is a
 * declaration reference that contains three member references:
 * `ui`, `.controls`, and `.Button`, and `.(render:static)`.
 */
export declare class DocMemberReference extends DocNode {
    /** {@inheritDoc} */
    readonly kind: DocNodeKind;
    private _dotParticle;
    private _leftParenthesisParticle;
    private _memberIdentifier;
    private _memberSymbol;
    private _colonParticle;
    private _selector;
    private _rightParenthesisParticle;
    /**
     * Don't call this directly.  Instead use {@link TSDocParser}
     * @internal
     */
    constructor(parameters: IDocMemberReferenceParameters);
    /**
     * True if this member reference is preceded by a dot (".") token.
     * It should be false only for the first member in the chain.
     */
    readonly hasDot: boolean;
    /**
     * The identifier for the referenced member.
     * @remarks
     * Either `memberIdentifier` or `memberSymbol` may be specified, but not both.
     */
    readonly memberIdentifier: DocMemberIdentifier | undefined;
    /**
     * The ECMAScript 6 symbol expression, which may be used instead of an identifier
     * to indicate the referenced member.
     * @remarks
     * Either `memberIdentifier` or `memberSymbol` may be specified, but not both.
     */
    readonly memberSymbol: DocMemberSymbol | undefined;
    /**
     * A TSDoc selector, which may be optionally when the identifier or symbol is insufficient
     * to unambiguously determine the referenced declaration.
     */
    readonly selector: DocMemberSelector | undefined;
    /** @override */
    updateParameters(parameters: IDocMemberReferenceParameters): void;
    /**
     * {@inheritDoc}
     * @override
     */
    getChildNodes(): ReadonlyArray<DocNode>;
}
