"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var Token_1 = require("./Token");
var Tokenizer_1 = require("./Tokenizer");
var nodes_1 = require("../nodes");
var TokenSequence_1 = require("./TokenSequence");
var Excerpt_1 = require("./Excerpt");
var TokenReader_1 = require("./TokenReader");
var StringChecks_1 = require("./StringChecks");
var TSDocTagDefinition_1 = require("./TSDocTagDefinition");
var StandardTags_1 = require("../details/StandardTags");
var PlainTextRenderer_1 = require("../renderers/PlainTextRenderer");
function isFailure(resultOrFailure) {
    return resultOrFailure !== undefined && resultOrFailure.hasOwnProperty('failureMessage');
}
/**
 * The main parser for TSDoc comments.
 */
var NodeParser = /** @class */ (function () {
    function NodeParser(parserContext) {
        this._parserContext = parserContext;
        this._currentSection = parserContext.docComment.summarySection;
    }
    NodeParser.prototype.parse = function () {
        var tokenReader = new TokenReader_1.TokenReader(this._parserContext);
        var done = false;
        while (!done) {
            // Extract the next token
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.EndOfInput:
                    done = true;
                    break;
                case Token_1.TokenKind.Newline:
                    this._pushAccumulatedPlainText(tokenReader);
                    tokenReader.readToken();
                    this._pushParagraphNode(new nodes_1.DocSoftBreak({
                        excerpt: new Excerpt_1.Excerpt({ content: tokenReader.extractAccumulatedSequence() })
                    }));
                    break;
                case Token_1.TokenKind.Backslash:
                    this._pushAccumulatedPlainText(tokenReader);
                    this._pushParagraphNode(this._parseBackslashEscape(tokenReader));
                    break;
                case Token_1.TokenKind.AtSign:
                    this._pushAccumulatedPlainText(tokenReader);
                    this._parseAndPushBlock(tokenReader);
                    break;
                case Token_1.TokenKind.LeftCurlyBracket: {
                    this._pushAccumulatedPlainText(tokenReader);
                    var marker = tokenReader.createMarker();
                    var docNode = this._parseInlineTag(tokenReader);
                    var docComment = this._parserContext.docComment;
                    if (docNode instanceof nodes_1.DocInheritDocTag) {
                        // The @inheritDoc tag is irregular because it looks like an inline tag, but
                        // it actually represents the entire comment body
                        var tagEndMarker = tokenReader.createMarker() - 1;
                        if (docComment.inheritDocTag === undefined) {
                            this._parserContext.docComment.inheritDocTag = docNode;
                        }
                        else {
                            this._pushParagraphNode(this._backtrackAndCreateErrorRange(tokenReader, marker, tagEndMarker, 'A doc comment cannot have more than one @inheritDoc tag'));
                        }
                    }
                    else {
                        this._pushParagraphNode(docNode);
                    }
                    break;
                }
                case Token_1.TokenKind.RightCurlyBracket:
                    this._pushAccumulatedPlainText(tokenReader);
                    this._pushParagraphNode(this._createError(tokenReader, 'The "}" character should be escaped using a backslash to avoid confusion with a TSDoc inline tag'));
                    break;
                case Token_1.TokenKind.LessThan:
                    this._pushAccumulatedPlainText(tokenReader);
                    // Look ahead two tokens to see if this is "<a>" or "</a>".
                    if (tokenReader.peekTokenAfterKind() === Token_1.TokenKind.Slash) {
                        this._pushParagraphNode(this._parseHtmlEndTag(tokenReader));
                    }
                    else {
                        this._pushParagraphNode(this._parseHtmlStartTag(tokenReader));
                    }
                    break;
                case Token_1.TokenKind.GreaterThan:
                    this._pushAccumulatedPlainText(tokenReader);
                    this._pushParagraphNode(this._createError(tokenReader, 'The ">" character should be escaped using a backslash to avoid confusion with an HTML tag'));
                    break;
                case Token_1.TokenKind.Backtick:
                    this._pushAccumulatedPlainText(tokenReader);
                    if (tokenReader.peekTokenAfterKind() === Token_1.TokenKind.Backtick
                        && tokenReader.peekTokenAfterAfterKind() === Token_1.TokenKind.Backtick) {
                        this._pushSectionNode(this._parseCodeFence(tokenReader));
                    }
                    else {
                        this._pushParagraphNode(this._parseCodeSpan(tokenReader));
                    }
                    break;
                default:
                    // If nobody recognized this token, then accumulate plain text
                    tokenReader.readToken();
                    break;
            }
        }
        this._pushAccumulatedPlainText(tokenReader);
        this._performValidationChecks();
    };
    NodeParser.prototype._performValidationChecks = function () {
        var docComment = this._parserContext.docComment;
        if (docComment.deprecatedBlock) {
            if (!PlainTextRenderer_1.PlainTextRenderer.hasAnyTextContent(docComment.deprecatedBlock)) {
                this._parserContext.log.addMessageForTokenSequence("The " + docComment.deprecatedBlock.blockTag.tagName + " block must include a deprecation message,"
                    + " e.g. describing the recommended alternative", docComment.deprecatedBlock.blockTag.excerpt.content, docComment.deprecatedBlock);
            }
        }
        if (docComment.inheritDocTag) {
            if (docComment.remarksBlock) {
                this._parserContext.log.addMessageForTokenSequence("A \"" + docComment.remarksBlock.blockTag.tagName + "\" block must not be used, because that"
                    + " content is provided by the @inheritDoc tag", docComment.remarksBlock.blockTag.excerpt.content, docComment.remarksBlock.blockTag);
            }
            if (PlainTextRenderer_1.PlainTextRenderer.hasAnyTextContent(docComment.summarySection)) {
                this._parserContext.log.addMessageForTextRange('The summary section must not have any content, because that'
                    + ' content is provided by the @inheritDoc tag', this._parserContext.commentRange);
            }
        }
    };
    NodeParser.prototype._validateTagDefinition = function (tagDefinition, tagName, expectingInlineTag, tokenSequenceForErrorContext, nodeForErrorContext) {
        if (tagDefinition) {
            var isInlineTag = tagDefinition.syntaxKind === TSDocTagDefinition_1.TSDocTagSyntaxKind.InlineTag;
            if (isInlineTag !== expectingInlineTag) {
                // The tag is defined, but it is used incorrectly
                if (expectingInlineTag) {
                    this._parserContext.log.addMessageForTokenSequence("The TSDoc tag \"" + tagName + "\" is an inline tag; it must be enclosed in \"{ }\" braces", tokenSequenceForErrorContext, nodeForErrorContext);
                }
                else {
                    this._parserContext.log.addMessageForTokenSequence("The TSDoc tag \"" + tagName + "\" is not an inline tag; it must not be enclosed in \"{ }\" braces", tokenSequenceForErrorContext, nodeForErrorContext);
                }
            }
            else {
                if (this._parserContext.configuration.validation.reportUnsupportedTags) {
                    if (!this._parserContext.configuration.isTagSupported(tagDefinition)) {
                        // The tag is defined, but not supported
                        this._parserContext.log.addMessageForTokenSequence("The TSDoc tag \"" + tagName + "\" is not supported by this tool", tokenSequenceForErrorContext, nodeForErrorContext);
                    }
                }
            }
        }
        else {
            // The tag is not defined
            if (!this._parserContext.configuration.validation.ignoreUndefinedTags) {
                this._parserContext.log.addMessageForTokenSequence("The TSDoc tag \"" + tagName + "\" is not defined in this configuration", tokenSequenceForErrorContext, nodeForErrorContext);
            }
        }
    };
    NodeParser.prototype._pushAccumulatedPlainText = function (tokenReader) {
        if (!tokenReader.isAccumulatedSequenceEmpty()) {
            var plainTextSequence = tokenReader.extractAccumulatedSequence();
            this._pushParagraphNode(new nodes_1.DocPlainText({
                text: plainTextSequence.toString(),
                excerpt: new Excerpt_1.Excerpt({ content: plainTextSequence })
            }));
        }
    };
    NodeParser.prototype._parseAndPushBlock = function (tokenReader) {
        var docComment = this._parserContext.docComment;
        var configuration = this._parserContext.configuration;
        var modifierTagSet = docComment.modifierTagSet;
        var parsedBlockTag = this._parseBlockTag(tokenReader);
        if (parsedBlockTag.kind !== "BlockTag" /* BlockTag */) {
            this._pushParagraphNode(parsedBlockTag);
            return;
        }
        var docBlockTag = parsedBlockTag;
        // Do we have a definition for this tag?
        var tagDefinition = configuration.tryGetTagDefinitionWithUpperCase(docBlockTag.tagNameWithUpperCase);
        this._validateTagDefinition(tagDefinition, docBlockTag.tagName, /* expectingInlineTag */ false, docBlockTag.excerpt.content, docBlockTag);
        if (tagDefinition) {
            switch (tagDefinition.syntaxKind) {
                case TSDocTagDefinition_1.TSDocTagSyntaxKind.BlockTag:
                    if (docBlockTag.tagNameWithUpperCase === StandardTags_1.StandardTags.param.tagNameWithUpperCase) {
                        var docParamBlock = this._parseParamBlock(tokenReader, docBlockTag);
                        this._parserContext.docComment.paramBlocks.push(docParamBlock);
                        this._currentSection = docParamBlock;
                        return;
                    }
                    else if (docBlockTag.tagNameWithUpperCase === StandardTags_1.StandardTags.typeParam.tagNameWithUpperCase) {
                        var docParamBlock = this._parseParamBlock(tokenReader, docBlockTag);
                        this._parserContext.docComment.typeParamBlocks.push(docParamBlock);
                        this._currentSection = docParamBlock;
                        return;
                    }
                    else {
                        var newBlock = new nodes_1.DocBlock({
                            blockTag: docBlockTag
                        });
                        this._addBlockToDocComment(newBlock);
                        this._currentSection = newBlock;
                    }
                    return;
                case TSDocTagDefinition_1.TSDocTagSyntaxKind.ModifierTag:
                    // The block tag was recognized as a modifier, so add it to the modifier tag set
                    // and do NOT call currentSection.appendNode(parsedNode)
                    modifierTagSet.addTag(docBlockTag);
                    return;
            }
        }
        this._pushParagraphNode(docBlockTag);
    };
    NodeParser.prototype._addBlockToDocComment = function (block) {
        var docComment = this._parserContext.docComment;
        switch (block.blockTag.tagNameWithUpperCase) {
            case StandardTags_1.StandardTags.remarks.tagNameWithUpperCase:
                docComment.remarksBlock = block;
                break;
            case StandardTags_1.StandardTags.privateRemarks.tagNameWithUpperCase:
                docComment.privateRemarks = block;
                break;
            case StandardTags_1.StandardTags.deprecated.tagNameWithUpperCase:
                docComment.deprecatedBlock = block;
                break;
            case StandardTags_1.StandardTags.returns.tagNameWithUpperCase:
                docComment.returnsBlock = block;
                break;
            default:
                docComment.appendCustomBlock(block);
        }
    };
    NodeParser.prototype._parseParamBlock = function (tokenReader, docBlockTag) {
        var startMarker = tokenReader.createMarker();
        this._readSpacingAndNewlines(tokenReader);
        var leadingWhitespaceSequence = tokenReader.tryExtractAccumulatedSequence();
        var parameterName = '';
        var done = false;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.AsciiWord:
                case Token_1.TokenKind.Period:
                    parameterName += tokenReader.readToken();
                    break;
                default:
                    done = true;
                    break;
            }
        }
        if (parameterName.length === 0) {
            tokenReader.backtrackToMarker(startMarker);
            var errorParamBlock = new nodes_1.DocParamBlock({
                blockTag: docBlockTag,
                parameterName: ''
            });
            this._parserContext.log.addMessageForTokenSequence('The @param block should be followed by a parameter name', docBlockTag.excerpt.content, docBlockTag);
            return errorParamBlock;
        }
        var parameterNameExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        // TODO: Warn if there is no space before or after the hyphen
        this._readSpacingAndNewlines(tokenReader);
        parameterNameExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Hyphen) {
            tokenReader.backtrackToMarker(startMarker);
            this._parserContext.log.addMessageForTokenSequence('The @param block should be followed by a parameter name and then a hyphen', docBlockTag.excerpt.content, docBlockTag);
            return new nodes_1.DocParamBlock({
                blockTag: docBlockTag,
                parameterName: ''
            });
        }
        tokenReader.readToken();
        var hyphenExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        // TODO: Only read one space
        this._readSpacingAndNewlines(tokenReader);
        hyphenExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        if (leadingWhitespaceSequence) {
            // The leading whitespace that we parsed to the docBlockTag
            docBlockTag.updateExcerpt(new Excerpt_1.Excerpt({
                content: docBlockTag.excerpt.content,
                spacingAfterContent: leadingWhitespaceSequence
            }));
        }
        return new nodes_1.DocParamBlock({
            blockTag: docBlockTag,
            parameterNameExcerpt: new Excerpt_1.Excerpt(parameterNameExcerptParameters),
            parameterName: parameterName,
            hyphenExcerpt: new Excerpt_1.Excerpt(hyphenExcerptParameters)
        });
    };
    NodeParser.prototype._pushParagraphNode = function (docNode) {
        this._currentSection.appendNodeInParagraph(docNode);
    };
    NodeParser.prototype._pushSectionNode = function (docNode) {
        this._currentSection.appendNode(docNode);
    };
    NodeParser.prototype._parseBackslashEscape = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var marker = tokenReader.createMarker();
        tokenReader.readToken();
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.EndOfInput) {
            return this._backtrackAndCreateError(tokenReader, marker, 'A backslash must precede another character that is being escaped');
        }
        var escapedToken = tokenReader.readToken();
        // In CommonMark, a backslash is only allowed before a punctuation
        // character.  In all other contexts, the backslash is interpreted as a
        // literal character.
        if (!Tokenizer_1.Tokenizer.isPunctuation(escapedToken.kind)) {
            return this._backtrackAndCreateError(tokenReader, marker, 'A backslash can only be used to escape a punctuation character');
        }
        var tokenSequence = tokenReader.extractAccumulatedSequence();
        return new nodes_1.DocEscapedText({
            excerpt: new Excerpt_1.Excerpt({ content: tokenSequence }),
            escapeStyle: nodes_1.EscapeStyle.CommonMarkBackslash,
            text: escapedToken.toString()
        });
    };
    NodeParser.prototype._parseBlockTag = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var marker = tokenReader.createMarker();
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.AtSign) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting a TSDoc tag starting with "@"');
        }
        // "@one" is a valid TSDoc tag at the start of a line, but "@one@two" is
        // a syntax error.  For two tags it should be "@one @two", or for literal text it
        // should be "\@one\@two".
        switch (tokenReader.peekPreviousTokenKind()) {
            case Token_1.TokenKind.EndOfInput:
            case Token_1.TokenKind.Spacing:
            case Token_1.TokenKind.Newline:
                break;
            default:
                return this._backtrackAndCreateError(tokenReader, marker, 'A TSDoc tag must be preceded by whitespace');
        }
        // Include the "@" as part of the tagName
        var tagName = tokenReader.readToken().toString();
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.AsciiWord) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting a TSDoc tag name after the "@" character (or use a backslash to escape this character)');
        }
        var tagNameMarker = tokenReader.createMarker();
        while (tokenReader.peekTokenKind() === Token_1.TokenKind.AsciiWord) {
            tagName += tokenReader.readToken().toString();
        }
        if (tagName.length === 0) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting an inline TSDoc tag name immediately after "{@"');
        }
        if (StringChecks_1.StringChecks.explainIfInvalidTSDocTagName(tagName)) {
            var failure = this._createFailureForTokensSince(tokenReader, 'A TSDoc tag name must start with a letter and contain only letters and numbers', tagNameMarker);
            return this._backtrackAndCreateErrorForFailure(tokenReader, marker, '', failure);
        }
        switch (tokenReader.peekTokenKind()) {
            case Token_1.TokenKind.Spacing:
            case Token_1.TokenKind.Newline:
            case Token_1.TokenKind.EndOfInput:
                break;
            default:
                return this._backtrackAndCreateError(tokenReader, marker, 'A TSDoc tag must be followed by whitespace');
        }
        return new nodes_1.DocBlockTag({
            excerpt: new Excerpt_1.Excerpt({ content: tokenReader.extractAccumulatedSequence() }),
            tagName: tagName
        });
    };
    NodeParser.prototype._parseInlineTag = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var marker = tokenReader.createMarker();
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.LeftCurlyBracket) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting a TSDoc tag starting with "{"');
        }
        tokenReader.readToken();
        var openingDelimiterExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        // For inline tags, if we handle errors by backtracking to the "{"  token, then the main loop
        // will then interpret the "@" as a block tag, which is almost certainly incorrect.  So the
        // DocErrorText needs to include both the "{" and "@" tokens.
        // We will use _backtrackAndCreateErrorRangeForFailure() for that.
        var atSignMarker = tokenReader.createMarker();
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.AtSign) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting a TSDoc tag starting with "{@"');
        }
        // Include the "@" as part of the tagName
        var tagName = tokenReader.readToken().toString();
        while (tokenReader.peekTokenKind() === Token_1.TokenKind.AsciiWord) {
            tagName += tokenReader.readToken().toString();
        }
        if (tagName === '@') {
            // This is an unusual case
            var failure = this._createFailureForTokensSince(tokenReader, 'Expecting a TSDoc inline tag name after the "{@" characters', atSignMarker);
            return this._backtrackAndCreateErrorRangeForFailure(tokenReader, marker, atSignMarker, '', failure);
        }
        if (StringChecks_1.StringChecks.explainIfInvalidTSDocTagName(tagName)) {
            var failure = this._createFailureForTokensSince(tokenReader, 'A TSDoc tag name must start with a letter and contain only letters and numbers', atSignMarker);
            return this._backtrackAndCreateErrorRangeForFailure(tokenReader, marker, atSignMarker, '', failure);
        }
        var tagNameExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        var spacingAfterTagName = this._readSpacingAndNewlines(tokenReader);
        tagNameExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        if (spacingAfterTagName.length === 0) {
            // If there were no spaces at all, that's an error unless it's the degenerate "{@tag}" case
            if (tokenReader.peekTokenKind() !== Token_1.TokenKind.RightCurlyBracket) {
                var failure = this._createFailureForToken(tokenReader, 'Expecting a space after the TSDoc inline tag name');
                return this._backtrackAndCreateErrorRangeForFailure(tokenReader, marker, atSignMarker, '', failure);
            }
        }
        var tagContent = '';
        var done = false;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.EndOfInput:
                    return this._backtrackAndCreateErrorRange(tokenReader, marker, atSignMarker, 'The TSDoc inline tag name is missing its closing "}"');
                case Token_1.TokenKind.Backslash:
                    // http://usejsdoc.org/about-block-inline-tags.html
                    // "If your tag's text includes a closing curly brace (}), you must escape it with
                    // a leading backslash (\)."
                    tokenReader.readToken(); // discard the backslash
                    // In CommonMark, a backslash is only allowed before a punctuation
                    // character.  In all other contexts, the backslash is interpreted as a
                    // literal character.
                    if (!Tokenizer_1.Tokenizer.isPunctuation(tokenReader.peekTokenKind())) {
                        var failure = this._createFailureForToken(tokenReader, 'A backslash can only be used to escape a punctuation character');
                        return this._backtrackAndCreateErrorRangeForFailure(tokenReader, marker, atSignMarker, 'Error reading inline TSDoc tag: ', failure);
                    }
                    tagContent += tokenReader.readToken().toString();
                    break;
                case Token_1.TokenKind.LeftCurlyBracket:
                    {
                        var failure = this._createFailureForToken(tokenReader, 'The "{" character must be escaped with a backslash when used inside a TSDoc inline tag');
                        return this._backtrackAndCreateErrorRangeForFailure(tokenReader, marker, atSignMarker, '', failure);
                    }
                case Token_1.TokenKind.RightCurlyBracket:
                    done = true;
                    break;
                default:
                    tagContent += tokenReader.readToken().toString();
                    break;
            }
        }
        var tagContentExcerpt;
        if (!tokenReader.isAccumulatedSequenceEmpty()) {
            tagContentExcerpt = new Excerpt_1.Excerpt({
                content: tokenReader.extractAccumulatedSequence()
            });
        }
        // Read the right curly bracket
        tokenReader.readToken();
        var closingDelimiterExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        var docInlineTagParameters = {
            openingDelimiterExcerpt: new Excerpt_1.Excerpt(openingDelimiterExcerptParameters),
            tagNameExcerpt: new Excerpt_1.Excerpt(tagNameExcerptParameters),
            tagName: tagName,
            tagContentExcerpt: tagContentExcerpt,
            tagContent: tagContent,
            closingDelimiterExcerpt: new Excerpt_1.Excerpt(closingDelimiterExcerptParameters)
        };
        var tagNameWithUpperCase = tagName.toUpperCase();
        // Create a new TokenReader that will reparse the tokens corresponding to the tagContent.
        var embeddedTokenReader = new TokenReader_1.TokenReader(this._parserContext, tagContentExcerpt ? tagContentExcerpt.content : TokenSequence_1.TokenSequence.createEmpty(this._parserContext));
        var docNode;
        switch (tagNameWithUpperCase) {
            case StandardTags_1.StandardTags.inheritDoc.tagNameWithUpperCase:
                docNode = this._parseInheritDocTag(docInlineTagParameters, embeddedTokenReader);
                break;
            case StandardTags_1.StandardTags.link.tagNameWithUpperCase:
                docNode = this._parseLinkTag(docInlineTagParameters, embeddedTokenReader);
                break;
            default:
                docNode = new nodes_1.DocInlineTag(docInlineTagParameters);
        }
        // Validate the tag
        var tagDefinition = this._parserContext.configuration.tryGetTagDefinitionWithUpperCase(tagNameWithUpperCase);
        this._validateTagDefinition(tagDefinition, tagName, /* expectingInlineTag */ true, tagNameExcerptParameters.content, docNode);
        return docNode;
    };
    NodeParser.prototype._parseInheritDocTag = function (docInlineTagParameters, embeddedTokenReader) {
        var docInheritDocTag = new nodes_1.DocInheritDocTag(docInlineTagParameters);
        var parameters = __assign({}, docInlineTagParameters);
        if (embeddedTokenReader.peekTokenKind() !== Token_1.TokenKind.EndOfInput) {
            parameters.declarationReference = this._parseDeclarationReference(embeddedTokenReader, docInlineTagParameters.tagNameExcerpt.content, docInheritDocTag);
            if (!parameters.declarationReference) {
                return docInheritDocTag; // error
            }
            if (embeddedTokenReader.peekTokenKind() !== Token_1.TokenKind.EndOfInput) {
                embeddedTokenReader.readToken();
                this._parserContext.log.addMessageForTokenSequence('Unexpected character after declaration reference', embeddedTokenReader.extractAccumulatedSequence(), docInheritDocTag);
                return docInheritDocTag; // error
            }
        }
        // We don't need the tagContentExcerpt since those tokens are now associated with the link particles
        parameters.tagContentExcerpt = undefined;
        docInheritDocTag.updateParameters(parameters);
        return docInheritDocTag;
    };
    NodeParser.prototype._parseLinkTag = function (docInlineTagParameters, embeddedTokenReader) {
        var docLinkTag = new nodes_1.DocLinkTag(docInlineTagParameters);
        var parameters = __assign({}, docInlineTagParameters);
        if (!parameters.tagContentExcerpt) {
            this._parserContext.log.addMessageForTokenSequence('The @link tag content is missing', parameters.tagNameExcerpt.content, docLinkTag);
            return docLinkTag; // error
        }
        // Is the link destination a URL or a declaration reference?
        //
        // The JSDoc "@link" tag allows URLs, however supporting full URLs would be highly
        // ambiguous, for example "microsoft.windows.camera:" is an actual valid URI scheme,
        // and even the common "mailto:example.com" looks suspiciously like a declaration reference.
        // In practice JSDoc URLs are nearly always HTTP or HTTPS, so it seems fairly reasonable to
        // require the URL to have "://" and a scheme without any punctuation in it.  If a more exotic
        // URL is needed, the HTML "<a>" tag can always be used.
        // We start with a fairly broad classifier heuristic, and then the parsers will refine this:
        // 1. Does it start with "//"?
        // 2. Does it contain "://"?
        var looksLikeUrl = embeddedTokenReader.peekTokenKind() === Token_1.TokenKind.Slash
            && embeddedTokenReader.peekTokenAfterKind() === Token_1.TokenKind.Slash;
        var marker = embeddedTokenReader.createMarker();
        var done = looksLikeUrl;
        while (!done) {
            switch (embeddedTokenReader.peekTokenKind()) {
                // An URI scheme can contain letters, numbers, minus, plus, and periods
                case Token_1.TokenKind.AsciiWord:
                case Token_1.TokenKind.Period:
                case Token_1.TokenKind.Hyphen:
                case Token_1.TokenKind.Plus:
                    embeddedTokenReader.readToken();
                    break;
                case Token_1.TokenKind.Colon:
                    embeddedTokenReader.readToken();
                    // Once we a reach a colon, then it's a URL only if we see "://"
                    looksLikeUrl = embeddedTokenReader.peekTokenKind() === Token_1.TokenKind.Slash
                        && embeddedTokenReader.peekTokenAfterKind() === Token_1.TokenKind.Slash;
                    done = true;
                    break;
                default:
                    done = true;
            }
        }
        embeddedTokenReader.backtrackToMarker(marker);
        // Is the hyperlink a URL or a declaration reference?
        if (looksLikeUrl) {
            // It starts with something like "http://", so parse it as a URL
            if (!this._parseLinkTagUrlDestination(embeddedTokenReader, parameters, docInlineTagParameters.tagNameExcerpt.content, docLinkTag)) {
                return docLinkTag; // error
            }
        }
        else {
            // Otherwise, assume it's a declaration reference
            if (!this._parseLinkTagCodeDestination(embeddedTokenReader, parameters, docInlineTagParameters.tagNameExcerpt.content, docLinkTag)) {
                return docLinkTag; // error
            }
        }
        if (embeddedTokenReader.peekTokenKind() === Token_1.TokenKind.Spacing) {
            // The above parser rules should have consumed any spacing before the pipe
            throw new Error('Unconsumed spacing encountered after construct');
        }
        if (embeddedTokenReader.peekTokenKind() === Token_1.TokenKind.Pipe) {
            // Read the link text
            embeddedTokenReader.readToken();
            parameters.pipeExcerpt = new Excerpt_1.Excerpt({
                content: embeddedTokenReader.extractAccumulatedSequence()
            });
            // Read everything until the end
            // NOTE: Because we're using an embedded TokenReader, the TokenKind.EndOfInput occurs
            // when we reach the "}", not the end of the original input
            done = false;
            while (!done) {
                switch (embeddedTokenReader.peekTokenKind()) {
                    case Token_1.TokenKind.EndOfInput:
                        done = true;
                        break;
                    case Token_1.TokenKind.Pipe:
                    case Token_1.TokenKind.LeftCurlyBracket:
                        var badCharacter = embeddedTokenReader.readToken().toString();
                        this._parserContext.log.addMessageForTokenSequence("The \"" + badCharacter + "\" character may not be used in the link text without escaping it", embeddedTokenReader.extractAccumulatedSequence(), docLinkTag);
                        return docLinkTag; // error
                    default:
                        embeddedTokenReader.readToken();
                }
            }
            if (!embeddedTokenReader.isAccumulatedSequenceEmpty()) {
                var linkTextExcerpt = new Excerpt_1.Excerpt({
                    content: embeddedTokenReader.extractAccumulatedSequence()
                });
                parameters.linkText = linkTextExcerpt.content.toString();
                parameters.linkTextExcerpt = linkTextExcerpt;
            }
        }
        else if (embeddedTokenReader.peekTokenKind() !== Token_1.TokenKind.EndOfInput) {
            embeddedTokenReader.readToken();
            this._parserContext.log.addMessageForTokenSequence('Unexpected character after link destination', embeddedTokenReader.extractAccumulatedSequence(), docLinkTag);
            return docLinkTag; // error
        }
        // We don't need the tagContentExcerpt since those tokens are now associated with the link particles
        parameters.tagContentExcerpt = undefined;
        docLinkTag.updateParameters(parameters);
        return docLinkTag;
    };
    NodeParser.prototype._parseLinkTagUrlDestination = function (embeddedTokenReader, parameters, tokenSequenceForErrorContext, nodeForErrorContext) {
        // Simply accumulate everything up to the next space. We won't try to implement a proper
        // URI parser here.
        var url = '';
        var done = false;
        while (!done) {
            switch (embeddedTokenReader.peekTokenKind()) {
                case Token_1.TokenKind.Spacing:
                case Token_1.TokenKind.Newline:
                case Token_1.TokenKind.EndOfInput:
                case Token_1.TokenKind.Pipe:
                case Token_1.TokenKind.RightCurlyBracket:
                    done = true;
                    break;
                default:
                    url += embeddedTokenReader.readToken();
                    break;
            }
        }
        if (url.length === 0) {
            // This should be impossible since the caller ensures that peekTokenKind() === TokenKind.AsciiWord
            throw new Error('Missing URL in _parseLinkTagUrl()');
        }
        var excerptParameters = {
            content: embeddedTokenReader.extractAccumulatedSequence()
        };
        var invalidUrlExplanation = StringChecks_1.StringChecks.explainIfInvalidLinkUrl(url);
        if (invalidUrlExplanation) {
            this._parserContext.log.addMessageForTokenSequence(invalidUrlExplanation, excerptParameters.content, nodeForErrorContext);
            return false;
        }
        this._readSpacingAndNewlines(embeddedTokenReader);
        excerptParameters.spacingAfterContent = embeddedTokenReader.tryExtractAccumulatedSequence();
        parameters.urlDestination = url;
        parameters.urlDestinationExcerpt = new Excerpt_1.Excerpt(excerptParameters);
        return true;
    };
    NodeParser.prototype._parseLinkTagCodeDestination = function (embeddedTokenReader, parameters, tokenSequenceForErrorContext, nodeForErrorContext) {
        parameters.codeDestination = this._parseDeclarationReference(embeddedTokenReader, tokenSequenceForErrorContext, nodeForErrorContext);
        return !!parameters.codeDestination;
    };
    NodeParser.prototype._parseDeclarationReference = function (tokenReader, tokenSequenceForErrorContext, nodeForErrorContext) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        // The package name can contain characters that look like a member reference.  This means we need to scan forwards
        // to see if there is a "#".  However, we need to be careful not to match a "#" that is part of a quoted expression.
        var marker = tokenReader.createMarker();
        var hasHash = false;
        var done = false;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.DoubleQuote:
                case Token_1.TokenKind.EndOfInput:
                case Token_1.TokenKind.LeftCurlyBracket:
                case Token_1.TokenKind.LeftParenthesis:
                case Token_1.TokenKind.LeftSquareBracket:
                case Token_1.TokenKind.Newline:
                case Token_1.TokenKind.Pipe:
                case Token_1.TokenKind.RightCurlyBracket:
                case Token_1.TokenKind.RightParenthesis:
                case Token_1.TokenKind.RightSquareBracket:
                case Token_1.TokenKind.SingleQuote:
                case Token_1.TokenKind.Spacing:
                    done = true;
                    break;
                case Token_1.TokenKind.PoundSymbol:
                    hasHash = true;
                    done = true;
                    break;
                default:
                    tokenReader.readToken();
            }
        }
        tokenReader.backtrackToMarker(marker);
        var packageNameExcerpt;
        var importPathExcerpt;
        var importHashExcerpt;
        if (hasHash) {
            // If it starts with a "." then it's a relative path, not a package name
            if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Period) {
                // Read the package name:
                var scopedPackageName = tokenReader.peekTokenKind() === Token_1.TokenKind.AtSign;
                var finishedScope = false;
                done = false;
                while (!done) {
                    switch (tokenReader.peekTokenKind()) {
                        case Token_1.TokenKind.EndOfInput:
                            // If hasHash=true, then we are expecting to stop when we reach the hash
                            throw new Error('Expecting pound symbol');
                        case Token_1.TokenKind.Slash:
                            // Stop at the first slash, unless this is a scoped package, in which case we stop at the second slash
                            if (scopedPackageName && !finishedScope) {
                                tokenReader.readToken();
                                finishedScope = true;
                            }
                            else {
                                done = true;
                            }
                            break;
                        case Token_1.TokenKind.PoundSymbol:
                            done = true;
                            break;
                        default:
                            tokenReader.readToken();
                    }
                }
                if (!tokenReader.isAccumulatedSequenceEmpty()) {
                    var packageNameExcerptParameters = {
                        content: tokenReader.extractAccumulatedSequence()
                    };
                    this._readSpacingAndNewlines(tokenReader);
                    packageNameExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
                    packageNameExcerpt = new Excerpt_1.Excerpt(packageNameExcerptParameters);
                    // Check that the packageName is syntactically valid
                    var explanation = StringChecks_1.StringChecks.explainIfInvalidPackageName(packageNameExcerpt.content.toString());
                    if (explanation) {
                        this._parserContext.log.addMessageForTokenSequence(explanation, packageNameExcerpt.content, nodeForErrorContext);
                        return undefined;
                    }
                }
            }
            // Read the import path:
            done = false;
            while (!done) {
                switch (tokenReader.peekTokenKind()) {
                    case Token_1.TokenKind.EndOfInput:
                        // If hasHash=true, then we are expecting to stop when we reach the hash
                        throw new Error('Expecting pound symbol');
                    case Token_1.TokenKind.PoundSymbol:
                        done = true;
                        break;
                    default:
                        tokenReader.readToken();
                }
            }
            if (!tokenReader.isAccumulatedSequenceEmpty()) {
                var importPathExcerptParameters = {
                    content: tokenReader.extractAccumulatedSequence()
                };
                this._readSpacingAndNewlines(tokenReader);
                importPathExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
                importPathExcerpt = new Excerpt_1.Excerpt(importPathExcerptParameters);
                // Check that the importPath is syntactically valid
                var explanation = StringChecks_1.StringChecks.explainIfInvalidImportPath(importPathExcerpt.content.toString(), !!packageNameExcerpt);
                if (explanation) {
                    this._parserContext.log.addMessageForTokenSequence(explanation, importPathExcerpt.content, nodeForErrorContext);
                    return undefined;
                }
            }
            // Read the import hash
            if (tokenReader.peekTokenKind() !== Token_1.TokenKind.PoundSymbol) {
                // The above logic should have left us at the PoundSymbol
                throw new Error('Expecting pound symbol');
            }
            tokenReader.readToken();
            var importHashExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            importHashExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            importHashExcerpt = new Excerpt_1.Excerpt(importHashExcerptParameters);
            if (packageNameExcerpt === undefined && importPathExcerpt === undefined) {
                this._parserContext.log.addMessageForTokenSequence('The hash character must be preceded by a package name or import path', importHashExcerpt.content, nodeForErrorContext);
                return undefined;
            }
        }
        // Read the member references:
        var memberReferences = [];
        done = false;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.Period:
                case Token_1.TokenKind.LeftParenthesis:
                case Token_1.TokenKind.AsciiWord:
                case Token_1.TokenKind.Colon:
                case Token_1.TokenKind.LeftSquareBracket:
                case Token_1.TokenKind.DoubleQuote:
                    var expectingDot = memberReferences.length > 0;
                    var memberReference = this._parseMemberReference(tokenReader, expectingDot, tokenSequenceForErrorContext, nodeForErrorContext);
                    if (!memberReference) {
                        return undefined;
                    }
                    memberReferences.push(memberReference);
                    break;
                default:
                    done = true;
            }
        }
        if (packageNameExcerpt === undefined && importPathExcerpt === undefined && memberReferences.length === 0) {
            // We didn't find any parts of a declaration reference
            this._parserContext.log.addMessageForTokenSequence('Expecting a declaration reference', tokenSequenceForErrorContext, nodeForErrorContext);
            return undefined;
        }
        return new nodes_1.DocDeclarationReference({
            packageNameExcerpt: packageNameExcerpt,
            packageName: packageNameExcerpt !== undefined ? packageNameExcerpt.content.toString() : undefined,
            importPathExcerpt: importPathExcerpt,
            importPath: importPathExcerpt !== undefined ? importPathExcerpt.content.toString() : undefined,
            importHashExcerpt: importHashExcerpt,
            memberReferences: memberReferences
        });
    };
    NodeParser.prototype._parseMemberReference = function (tokenReader, expectingDot, tokenSequenceForErrorContext, nodeForErrorContext) {
        var dotExcerpt;
        // Read the dot operator
        if (expectingDot) {
            if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Period) {
                this._parserContext.log.addMessageForTokenSequence('Expecting a period before the next component'
                    + ' of a declaration reference', tokenSequenceForErrorContext, nodeForErrorContext);
                return undefined;
            }
            tokenReader.readToken();
            var dotExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            dotExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            dotExcerpt = new Excerpt_1.Excerpt(dotExcerptParameters);
        }
        // Read the left parenthesis if there is one
        var leftParenthesisExcerpt;
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.LeftParenthesis) {
            tokenReader.readToken();
            var leftParenthesisExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            leftParenthesisExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            leftParenthesisExcerpt = new Excerpt_1.Excerpt(leftParenthesisExcerptParameters);
        }
        var memberIdentifier = undefined;
        var memberSymbol = undefined;
        // Read the member identifier or symbol
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.LeftSquareBracket) {
            memberSymbol = this._parseMemberSymbol(tokenReader, nodeForErrorContext);
            if (!memberSymbol) {
                return undefined;
            }
        }
        else {
            memberIdentifier = this._parseMemberIdentifier(tokenReader, tokenSequenceForErrorContext, nodeForErrorContext);
            if (!memberIdentifier) {
                return undefined;
            }
        }
        // Read the colon
        var colonExcerpt;
        var selector = undefined;
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.Colon) {
            tokenReader.readToken();
            var colonExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            colonExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            colonExcerpt = new Excerpt_1.Excerpt(colonExcerptParameters);
            if (!leftParenthesisExcerpt) {
                // In the current TSDoc draft standard, a member reference with a selector requires the parentheses.
                // It would be reasonable to make the parentheses optional, and we are contemplating simplifying the
                // notation in the future.  But for now the parentheses are required.
                this._parserContext.log.addMessageForTokenSequence('Syntax error in declaration reference: the member selector must be enclosed in parentheses', colonExcerpt.content, nodeForErrorContext);
                return undefined;
            }
            // If there is a colon, then read the selector
            selector = this._parseMemberSelector(tokenReader, colonExcerptParameters.content, nodeForErrorContext);
            if (!selector) {
                return undefined;
            }
        }
        else {
            if (leftParenthesisExcerpt) {
                this._parserContext.log.addMessageForTokenSequence('Expecting a colon after the identifier because'
                    + ' the expression is in parentheses', leftParenthesisExcerpt.content, nodeForErrorContext);
                return undefined;
            }
        }
        // Read the right parenthesis
        var rightParenthesisExcerpt;
        if (leftParenthesisExcerpt) {
            if (tokenReader.peekTokenKind() !== Token_1.TokenKind.RightParenthesis) {
                this._parserContext.log.addMessageForTokenSequence('Expecting a matching right parenthesis', leftParenthesisExcerpt.content, nodeForErrorContext);
                return undefined;
            }
            tokenReader.readToken();
            var rightParenthesisExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            rightParenthesisExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            rightParenthesisExcerpt = new Excerpt_1.Excerpt(rightParenthesisExcerptParameters);
        }
        return new nodes_1.DocMemberReference({
            hasDot: dotExcerpt !== undefined,
            dotExcerpt: dotExcerpt,
            leftParenthesisExcerpt: leftParenthesisExcerpt,
            memberIdentifier: memberIdentifier,
            memberSymbol: memberSymbol,
            colonExcerpt: colonExcerpt,
            selector: selector,
            rightParenthesisExcerpt: rightParenthesisExcerpt
        });
    };
    NodeParser.prototype._parseMemberSymbol = function (tokenReader, nodeForErrorContext) {
        // Read the "["
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.LeftSquareBracket) {
            // This should be impossible since the caller ensures that peekTokenKind() === TokenKind.LeftSquareBracket
            throw new Error('Expecting "["');
        }
        tokenReader.readToken();
        var leftBracketExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        this._readSpacingAndNewlines(tokenReader);
        leftBracketExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        // Read the declaration reference
        var declarationReference = this._parseDeclarationReference(tokenReader, leftBracketExcerptParameters.content, nodeForErrorContext);
        if (!declarationReference) {
            this._parserContext.log.addMessageForTokenSequence('Missing declaration reference in symbol reference', leftBracketExcerptParameters.content, nodeForErrorContext);
            return undefined;
        }
        // Read the "]"
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.RightSquareBracket) {
            this._parserContext.log.addMessageForTokenSequence('Missing closing square bracket for symbol reference', leftBracketExcerptParameters.content, nodeForErrorContext);
            return undefined;
        }
        tokenReader.readToken();
        var rightBracketExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        this._readSpacingAndNewlines(tokenReader);
        rightBracketExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        return new nodes_1.DocMemberSymbol({
            leftBracketExcerpt: new Excerpt_1.Excerpt(leftBracketExcerptParameters),
            symbolReference: declarationReference,
            rightBracketExcerpt: new Excerpt_1.Excerpt(rightBracketExcerptParameters)
        });
    };
    NodeParser.prototype._parseMemberIdentifier = function (tokenReader, tokenSequenceForErrorContext, nodeForErrorContext) {
        // Is this a quoted identifier?
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.DoubleQuote) {
            // Read the opening '"'
            tokenReader.readToken();
            var leftQuoteExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            // Read the text inside the quotes
            while (tokenReader.peekTokenKind() !== Token_1.TokenKind.DoubleQuote) {
                if (tokenReader.peekTokenKind() === Token_1.TokenKind.EndOfInput) {
                    this._parserContext.log.addMessageForTokenSequence('Unexpected end of input inside quoted member identifier', leftQuoteExcerptParameters.content, nodeForErrorContext);
                    return undefined;
                }
                tokenReader.readToken();
            }
            if (tokenReader.isAccumulatedSequenceEmpty()) {
                this._parserContext.log.addMessageForTokenSequence('The quoted identifier cannot be empty', leftQuoteExcerptParameters.content, nodeForErrorContext);
                return undefined;
            }
            var identifierExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            // Read the closing '""
            tokenReader.readToken(); // read the quote
            var rightQuoteExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            rightQuoteExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            return new nodes_1.DocMemberIdentifier({
                leftQuoteExcerpt: new Excerpt_1.Excerpt(leftQuoteExcerptParameters),
                identifierExcerpt: new Excerpt_1.Excerpt(identifierExcerptParameters),
                identifier: identifierExcerptParameters.content.toString(),
                rightQuoteExcerpt: new Excerpt_1.Excerpt(rightQuoteExcerptParameters)
            });
        }
        else {
            // Otherwise assume it's a valid TypeScript identifier
            if (tokenReader.peekTokenKind() !== Token_1.TokenKind.AsciiWord) {
                this._parserContext.log.addMessageForTokenSequence('Syntax error in declaration reference: expecting a member identifier', tokenSequenceForErrorContext, nodeForErrorContext);
                return undefined;
            }
            var identifier = tokenReader.readToken().toString();
            var identifierExcerptParameters = {
                content: tokenReader.extractAccumulatedSequence()
            };
            this._readSpacingAndNewlines(tokenReader);
            identifierExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
            var explanation = StringChecks_1.StringChecks.explainIfInvalidUnquotedIdentifier(identifier);
            if (explanation) {
                this._parserContext.log.addMessageForTokenSequence(explanation, identifierExcerptParameters.content, nodeForErrorContext);
                return undefined;
            }
            return new nodes_1.DocMemberIdentifier({
                identifierExcerpt: new Excerpt_1.Excerpt(identifierExcerptParameters),
                identifier: identifier
            });
        }
    };
    NodeParser.prototype._parseMemberSelector = function (tokenReader, tokenSequenceForErrorContext, nodeForErrorContext) {
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.AsciiWord) {
            this._parserContext.log.addMessageForTokenSequence('Expecting a selector label after the colon', tokenSequenceForErrorContext, nodeForErrorContext);
        }
        var selector = tokenReader.readToken().toString();
        var selectorExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        this._readSpacingAndNewlines(tokenReader);
        selectorExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        var docMemberSelector = new nodes_1.DocMemberSelector({
            excerpt: new Excerpt_1.Excerpt(selectorExcerptParameters),
            selector: selector
        });
        if (docMemberSelector.errorMessage) {
            this._parserContext.log.addMessageForTokenSequence(docMemberSelector.errorMessage, selectorExcerptParameters.content, nodeForErrorContext);
            return undefined;
        }
        return docMemberSelector;
    };
    NodeParser.prototype._parseHtmlStartTag = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var marker = tokenReader.createMarker();
        // Read the "<" delimiter
        var lessThanToken = tokenReader.readToken();
        if (lessThanToken.kind !== Token_1.TokenKind.LessThan) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting an HTML tag starting with "<"');
        }
        // NOTE: CommonMark does not permit whitespace after the "<"
        var openingDelimiterExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        // Read the element name
        var elementName = this._parseHtmlName(tokenReader);
        if (isFailure(elementName)) {
            return this._backtrackAndCreateErrorForFailure(tokenReader, marker, 'Invalid HTML element: ', elementName);
        }
        var elementNameExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        var spacingAfterElementName = this._readSpacingAndNewlines(tokenReader);
        elementNameExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        var htmlAttributes = [];
        // Read the attributes until we see a ">" or "/>"
        while (tokenReader.peekTokenKind() === Token_1.TokenKind.AsciiWord) {
            // Read the attribute
            var attributeNode = this._parseHtmlAttribute(tokenReader);
            if (isFailure(attributeNode)) {
                return this._backtrackAndCreateErrorForFailure(tokenReader, marker, 'The HTML element has an invalid attribute: ', attributeNode);
            }
            htmlAttributes.push(attributeNode);
        }
        // Read the closing "/>" or ">" as the Excerpt.suffix
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var endDelimiterMarker = tokenReader.createMarker();
        var selfClosingTag = false;
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.Slash) {
            tokenReader.readToken();
            selfClosingTag = true;
        }
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.GreaterThan) {
            var failure = this._createFailureForTokensSince(tokenReader, 'Expecting an attribute or ">" or "/>"', endDelimiterMarker);
            return this._backtrackAndCreateErrorForFailure(tokenReader, marker, 'The HTML tag has invalid syntax: ', failure);
        }
        tokenReader.readToken();
        var closingDelimiterExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        // NOTE: We don't read excerptParameters.separator here, since if there is any it
        // will be represented as DocPlainText.
        return new nodes_1.DocHtmlStartTag({
            openingDelimiterExcerpt: new Excerpt_1.Excerpt(openingDelimiterExcerptParameters),
            elementNameExcerpt: new Excerpt_1.Excerpt(elementNameExcerptParameters),
            elementName: elementName,
            spacingAfterElementName: spacingAfterElementName,
            htmlAttributes: htmlAttributes,
            selfClosingTag: selfClosingTag,
            closingDelimiterExcerpt: new Excerpt_1.Excerpt(closingDelimiterExcerptParameters)
        });
    };
    NodeParser.prototype._parseHtmlAttribute = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        // Read the attribute name
        var attributeName = this._parseHtmlName(tokenReader);
        if (isFailure(attributeName)) {
            return attributeName;
        }
        var attributeNameExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        var spacingAfterAttributeName = this._readSpacingAndNewlines(tokenReader);
        attributeNameExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        // Read the equals
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Equals) {
            return this._createFailureForToken(tokenReader, 'Expecting "=" after HTML attribute name');
        }
        tokenReader.readToken();
        var equalsExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        var spacingAfterEquals = this._readSpacingAndNewlines(tokenReader);
        equalsExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        // Read the attribute value
        var attributeValue = this._parseHtmlString(tokenReader);
        if (isFailure(attributeValue)) {
            return attributeValue;
        }
        var attributeValueExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        var spacingAfterAttributeValue = this._readSpacingAndNewlines(tokenReader);
        attributeValueExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        return new nodes_1.DocHtmlAttribute({
            attributeNameExcerpt: new Excerpt_1.Excerpt(attributeNameExcerptParameters),
            attributeName: attributeName,
            spacingAfterAttributeName: spacingAfterAttributeName,
            equalsExcerpt: new Excerpt_1.Excerpt(equalsExcerptParameters),
            spacingAfterEquals: spacingAfterEquals,
            attributeValueExcerpt: new Excerpt_1.Excerpt(attributeValueExcerptParameters),
            attributeValue: attributeValue,
            spacingAfterAttributeValue: spacingAfterAttributeValue
        });
    };
    NodeParser.prototype._parseHtmlString = function (tokenReader) {
        var marker = tokenReader.createMarker();
        var quoteTokenKind = tokenReader.peekTokenKind();
        if (quoteTokenKind !== Token_1.TokenKind.DoubleQuote && quoteTokenKind !== Token_1.TokenKind.SingleQuote) {
            return this._createFailureForToken(tokenReader, 'Expecting an HTML string starting with a single-quote or double-quote character');
        }
        tokenReader.readToken();
        var textWithoutQuotes = '';
        while (true) {
            var peekedTokenKind = tokenReader.peekTokenKind();
            // Did we find the matching token?
            if (peekedTokenKind === quoteTokenKind) {
                tokenReader.readToken(); // extract the quote
                break;
            }
            if (peekedTokenKind === Token_1.TokenKind.EndOfInput || peekedTokenKind === Token_1.TokenKind.Newline) {
                return this._createFailureForToken(tokenReader, 'The HTML string is missing its closing quote', marker);
            }
            textWithoutQuotes += tokenReader.readToken().toString();
        }
        // The next attribute cannot start immediately after this one
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.AsciiWord) {
            return this._createFailureForToken(tokenReader, 'The next character after a closing quote must be spacing or punctuation');
        }
        return textWithoutQuotes;
    };
    NodeParser.prototype._parseHtmlEndTag = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var marker = tokenReader.createMarker();
        // Read the "</" delimiter
        var lessThanToken = tokenReader.peekToken();
        if (lessThanToken.kind !== Token_1.TokenKind.LessThan) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting an HTML tag starting with "</"');
        }
        tokenReader.readToken();
        var slashToken = tokenReader.peekToken();
        if (slashToken.kind !== Token_1.TokenKind.Slash) {
            return this._backtrackAndCreateError(tokenReader, marker, 'Expecting an HTML tag starting with "</"');
        }
        tokenReader.readToken();
        // NOTE: Spaces are not permitted here
        // https://www.w3.org/TR/html5/syntax.html#end-tags
        var openingDelimiterExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        // Read the tag name
        var elementName = this._parseHtmlName(tokenReader);
        if (isFailure(elementName)) {
            return this._backtrackAndCreateErrorForFailure(tokenReader, marker, 'Expecting an HTML element name: ', elementName);
        }
        var elementNameExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        this._readSpacingAndNewlines(tokenReader);
        elementNameExcerptParameters.spacingAfterContent = tokenReader.tryExtractAccumulatedSequence();
        // Read the closing ">"
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.GreaterThan) {
            var failure = this._createFailureForToken(tokenReader, 'Expecting a closing ">" for the HTML tag');
            return this._backtrackAndCreateErrorForFailure(tokenReader, marker, '', failure);
        }
        tokenReader.readToken();
        var closingDelimiterExcerptParameters = {
            content: tokenReader.extractAccumulatedSequence()
        };
        return new nodes_1.DocHtmlEndTag({
            openingDelimiterExcerpt: new Excerpt_1.Excerpt(openingDelimiterExcerptParameters),
            elementNameExcerpt: new Excerpt_1.Excerpt(elementNameExcerptParameters),
            elementName: elementName,
            closingDelimiterExcerpt: new Excerpt_1.Excerpt(closingDelimiterExcerptParameters)
        });
    };
    /**
     * Parses an HTML name such as an element name or attribute name.
     */
    NodeParser.prototype._parseHtmlName = function (tokenReader) {
        var htmlName = '';
        var marker = tokenReader.createMarker();
        if (tokenReader.peekTokenKind() === Token_1.TokenKind.Spacing) {
            return this._createFailureForTokensSince(tokenReader, 'A space is not allowed here', marker);
        }
        var done = false;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.AsciiWord:
                case Token_1.TokenKind.Hyphen:
                    htmlName += tokenReader.readToken().toString();
                    break;
                default:
                    done = true;
                    break;
            }
        }
        if (htmlName.length === 0) {
            return this._createFailureForToken(tokenReader, 'Expecting an HTML name');
        }
        if (!NodeParser.htmlNameRegExp.test(htmlName)) {
            return this._createFailureForTokensSince(tokenReader, 'An HTML name must be a sequence of letters separated by hyphens', marker);
        }
        return htmlName;
    };
    NodeParser.prototype._parseCodeFence = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var startMarker = tokenReader.createMarker();
        var endOfOpeningDelimiterMarker = startMarker + 2;
        switch (tokenReader.peekPreviousTokenKind()) {
            case Token_1.TokenKind.Newline:
            case Token_1.TokenKind.EndOfInput:
                break;
            default:
                return this._backtrackAndCreateErrorRange(tokenReader, startMarker, 
                // include the three backticks so they don't get reinterpreted as a code span
                endOfOpeningDelimiterMarker, 'The opening backtick for a code fence must appear at the start of the line');
        }
        // Read the opening ``` delimiter
        var openingDelimiter = '';
        openingDelimiter += tokenReader.readToken();
        openingDelimiter += tokenReader.readToken();
        openingDelimiter += tokenReader.readToken();
        if (openingDelimiter !== '```') {
            // This would be a parser bug -- the caller of _parseCodeFence() should have verified this while
            // looking ahead to distinguish code spans/fences
            throw new Error('Expecting three backticks');
        }
        var openingDelimiterSequence = tokenReader.extractAccumulatedSequence();
        // Read any spaces after the delimiter,
        // but NOT the Newline since that goes with the language particle
        while (tokenReader.peekTokenKind() === Token_1.TokenKind.Spacing) {
            tokenReader.readToken();
        }
        var openingDelimiterExcerpt = new Excerpt_1.Excerpt({
            content: openingDelimiterSequence,
            spacingAfterContent: tokenReader.tryExtractAccumulatedSequence()
        });
        // Read the language specifier (if present) and newline
        var done = false;
        var startOfPaddingMarker = undefined;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.Spacing:
                case Token_1.TokenKind.Newline:
                    if (startOfPaddingMarker === undefined) {
                        // Starting a new run of spacing characters
                        startOfPaddingMarker = tokenReader.createMarker();
                    }
                    if (tokenReader.peekTokenKind() === Token_1.TokenKind.Newline) {
                        done = true;
                    }
                    tokenReader.readToken();
                    break;
                case Token_1.TokenKind.Backtick:
                    var failure = this._createFailureForToken(tokenReader, 'The language specifier cannot contain backtick characters');
                    return this._backtrackAndCreateErrorRangeForFailure(tokenReader, startMarker, endOfOpeningDelimiterMarker, 'Error parsing code fence: ', failure);
                case Token_1.TokenKind.EndOfInput:
                    var failure2 = this._createFailureForToken(tokenReader, 'Missing closing delimiter');
                    return this._backtrackAndCreateErrorRangeForFailure(tokenReader, startMarker, endOfOpeningDelimiterMarker, 'Error parsing code fence: ', failure2);
                default:
                    // more non-spacing content
                    startOfPaddingMarker = undefined;
                    tokenReader.readToken();
                    break;
            }
        }
        // At this point, we must have accumulated at least a newline token.
        // Example: "pov-ray sdl    \n"
        var languageSequence = tokenReader.extractAccumulatedSequence();
        var languageExcerpt = new Excerpt_1.Excerpt({
            // Example: "pov-ray sdl"
            content: languageSequence.getNewSequence(languageSequence.startIndex, startOfPaddingMarker),
            // Example: "    \n"
            spacingAfterContent: languageSequence.getNewSequence(startOfPaddingMarker, languageSequence.endIndex)
        });
        // Read the code content until we see the closing ``` delimiter
        var codeEndMarker = -1;
        done = false;
        var tokenBeforeDelimiter;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.EndOfInput:
                    var failure2 = this._createFailureForToken(tokenReader, 'Missing closing delimiter');
                    return this._backtrackAndCreateErrorRangeForFailure(tokenReader, startMarker, endOfOpeningDelimiterMarker, 'Error parsing code fence: ', failure2);
                case Token_1.TokenKind.Newline:
                    tokenBeforeDelimiter = tokenReader.readToken();
                    codeEndMarker = tokenReader.createMarker();
                    while (tokenReader.peekTokenKind() === Token_1.TokenKind.Spacing) {
                        tokenBeforeDelimiter = tokenReader.readToken();
                    }
                    if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Backtick) {
                        break;
                    }
                    tokenReader.readToken(); // first backtick
                    if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Backtick) {
                        break;
                    }
                    tokenReader.readToken(); // second backtick
                    if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Backtick) {
                        break;
                    }
                    tokenReader.readToken(); // third backtick
                    done = true;
                    break;
                default:
                    tokenReader.readToken();
                    break;
            }
        }
        if (tokenBeforeDelimiter.kind !== Token_1.TokenKind.Newline) {
            this._parserContext.log.addMessageForTextRange('The closing delimiter for a code fence must not be indented', tokenBeforeDelimiter.range);
        }
        // Example: "code 1\ncode 2\n   ```"
        var codeAndDelimiterSequence = tokenReader.extractAccumulatedSequence();
        var codeExcerpt = new Excerpt_1.Excerpt({
            content: codeAndDelimiterSequence.getNewSequence(codeAndDelimiterSequence.startIndex, codeEndMarker)
        });
        // Read the spacing and newline after the closing delimiter
        done = false;
        while (!done) {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.Spacing:
                    tokenReader.readToken();
                    break;
                case Token_1.TokenKind.Newline:
                    done = true;
                    tokenReader.readToken();
                    break;
                case Token_1.TokenKind.EndOfInput:
                    done = true;
                    break;
                default:
                    this._parserContext.log.addMessageForTextRange('Unexpected characters after closing delimiter for code fence', tokenReader.peekToken().range);
                    done = true;
                    break;
            }
        }
        var closingDelimiterExcerpt = new Excerpt_1.Excerpt({
            // Example: "```"
            content: codeAndDelimiterSequence.getNewSequence(codeEndMarker, codeAndDelimiterSequence.endIndex),
            // Example: "   \n"
            spacingAfterContent: tokenReader.tryExtractAccumulatedSequence()
        });
        return new nodes_1.DocCodeFence({
            openingDelimiterExcerpt: openingDelimiterExcerpt,
            languageExcerpt: languageExcerpt,
            language: languageExcerpt.content.toString(),
            codeExcerpt: codeExcerpt,
            code: codeExcerpt.content.toString(),
            closingDelimiterExcerpt: closingDelimiterExcerpt
        });
    };
    NodeParser.prototype._parseCodeSpan = function (tokenReader) {
        tokenReader.assertAccumulatedSequenceIsEmpty();
        var marker = tokenReader.createMarker();
        // Parse the opening backtick
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.Backtick) {
            return this._createError(tokenReader, 'Expecting a code span starting with a backtick character "`"');
        }
        tokenReader.readToken(); // read the backtick
        var openingDelimiterExcerpt = new Excerpt_1.Excerpt({
            content: tokenReader.extractAccumulatedSequence()
        });
        var codeExcerpt;
        var closingDelimiterExcerpt;
        // Parse the content backtick
        while (true) {
            var peekedTokenKind = tokenReader.peekTokenKind();
            // Did we find the matching token?
            if (peekedTokenKind === Token_1.TokenKind.Backtick) {
                if (tokenReader.isAccumulatedSequenceEmpty()) {
                    return this._backtrackAndCreateErrorRange(tokenReader, marker, marker + 1, 'A code span must contain at least one character between the backticks');
                }
                codeExcerpt = new Excerpt_1.Excerpt({
                    content: tokenReader.extractAccumulatedSequence()
                });
                tokenReader.readToken();
                closingDelimiterExcerpt = new Excerpt_1.Excerpt({
                    content: tokenReader.extractAccumulatedSequence()
                });
                break;
            }
            if (peekedTokenKind === Token_1.TokenKind.EndOfInput || peekedTokenKind === Token_1.TokenKind.Newline) {
                return this._backtrackAndCreateError(tokenReader, marker, 'The code span is missing its closing backtick');
            }
            tokenReader.readToken();
        }
        return new nodes_1.DocCodeSpan({
            openingDelimiterExcerpt: openingDelimiterExcerpt,
            codeExcerpt: codeExcerpt,
            code: codeExcerpt.content.toString(),
            closingDelimiterExcerpt: closingDelimiterExcerpt
        });
    };
    NodeParser.prototype._readSpacingAndNewlines = function (tokenReader) {
        var result = '';
        var done = false;
        do {
            switch (tokenReader.peekTokenKind()) {
                case Token_1.TokenKind.Spacing:
                case Token_1.TokenKind.Newline:
                    result += tokenReader.readToken().toString();
                    break;
                default:
                    done = true;
                    break;
            }
        } while (!done);
        return result;
    };
    /**
     * Read the next token, and report it as a DocErrorText node.
     */
    NodeParser.prototype._createError = function (tokenReader, errorMessage) {
        tokenReader.readToken();
        var tokenSequence = tokenReader.extractAccumulatedSequence();
        var docErrorText = new nodes_1.DocErrorText({
            excerpt: new Excerpt_1.Excerpt({ content: tokenSequence }),
            text: tokenSequence.toString(),
            errorMessage: errorMessage,
            errorLocation: tokenSequence
        });
        this._parserContext.log.addMessageForDocErrorText(docErrorText);
        return docErrorText;
    };
    /**
     * Rewind to the specified marker, read the next token, and report it as a DocErrorText node.
     */
    NodeParser.prototype._backtrackAndCreateError = function (tokenReader, marker, errorMessage) {
        tokenReader.backtrackToMarker(marker);
        return this._createError(tokenReader, errorMessage);
    };
    /**
     * Rewind to the errorStartMarker, read the tokens up to and including errorInclusiveEndMarker,
     * and report it as a DocErrorText node.
     */
    NodeParser.prototype._backtrackAndCreateErrorRange = function (tokenReader, errorStartMarker, errorInclusiveEndMarker, errorMessage) {
        tokenReader.backtrackToMarker(errorStartMarker);
        while (tokenReader.createMarker() !== errorInclusiveEndMarker) {
            tokenReader.readToken();
        }
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.EndOfInput) {
            tokenReader.readToken();
        }
        var tokenSequence = tokenReader.extractAccumulatedSequence();
        var docErrorText = new nodes_1.DocErrorText({
            excerpt: new Excerpt_1.Excerpt({ content: tokenSequence }),
            text: tokenSequence.toString(),
            errorMessage: errorMessage,
            errorLocation: tokenSequence
        });
        this._parserContext.log.addMessageForDocErrorText(docErrorText);
        return docErrorText;
    };
    /**
     * Rewind to the specified marker, read the next token, and report it as a DocErrorText node
     * whose location is based on an IFailure.
     */
    NodeParser.prototype._backtrackAndCreateErrorForFailure = function (tokenReader, marker, errorMessagePrefix, failure) {
        tokenReader.backtrackToMarker(marker);
        tokenReader.readToken();
        var tokenSequence = tokenReader.extractAccumulatedSequence();
        var docErrorText = new nodes_1.DocErrorText({
            excerpt: new Excerpt_1.Excerpt({ content: tokenSequence }),
            text: tokenSequence.toString(),
            errorMessage: errorMessagePrefix + failure.failureMessage,
            errorLocation: failure.failureLocation
        });
        this._parserContext.log.addMessageForDocErrorText(docErrorText);
        return docErrorText;
    };
    /**
     * Rewind to the errorStartMarker, read the tokens up to and including errorInclusiveEndMarker,
     * and report it as a DocErrorText node whose location is based on an IFailure.
     */
    NodeParser.prototype._backtrackAndCreateErrorRangeForFailure = function (tokenReader, errorStartMarker, errorInclusiveEndMarker, errorMessagePrefix, failure) {
        tokenReader.backtrackToMarker(errorStartMarker);
        while (tokenReader.createMarker() !== errorInclusiveEndMarker) {
            tokenReader.readToken();
        }
        if (tokenReader.peekTokenKind() !== Token_1.TokenKind.EndOfInput) {
            tokenReader.readToken();
        }
        var tokenSequence = tokenReader.extractAccumulatedSequence();
        var docErrorText = new nodes_1.DocErrorText({
            excerpt: new Excerpt_1.Excerpt({ content: tokenSequence }),
            text: tokenSequence.toString(),
            errorMessage: errorMessagePrefix + failure.failureMessage,
            errorLocation: failure.failureLocation
        });
        this._parserContext.log.addMessageForDocErrorText(docErrorText);
        return docErrorText;
    };
    /**
     * Creates an IFailure whose TokenSequence is a single token.  If a marker is not specified,
     * then it is the current token.
     */
    NodeParser.prototype._createFailureForToken = function (tokenReader, failureMessage, tokenMarker) {
        if (!tokenMarker) {
            tokenMarker = tokenReader.createMarker();
        }
        var tokenSequence = new TokenSequence_1.TokenSequence({
            parserContext: this._parserContext,
            startIndex: tokenMarker,
            endIndex: tokenMarker + 1
        });
        return {
            failureMessage: failureMessage,
            failureLocation: tokenSequence
        };
    };
    /**
     * Creates an IFailure whose TokenSequence starts from the specified marker and
     * encompasses all tokens read since then.  If none were read, then the next token used.
     */
    NodeParser.prototype._createFailureForTokensSince = function (tokenReader, failureMessage, startMarker) {
        var endMarker = tokenReader.createMarker();
        if (endMarker < startMarker) {
            // This would be a parser bug
            throw new Error('Invalid startMarker');
        }
        if (endMarker === startMarker) {
            ++endMarker;
        }
        var tokenSequence = new TokenSequence_1.TokenSequence({
            parserContext: this._parserContext,
            startIndex: startMarker,
            endIndex: endMarker
        });
        return {
            failureMessage: failureMessage,
            failureLocation: tokenSequence
        };
    };
    // https://www.w3.org/TR/html5/syntax.html#tag-name
    // https://html.spec.whatwg.org/multipage/custom-elements.html#valid-custom-element-name
    NodeParser.htmlNameRegExp = /^[a-z]+(\-[a-z]+)*$/i;
    return NodeParser;
}());
exports.NodeParser = NodeParser;
//# sourceMappingURL=NodeParser.js.map