"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var nodes_1 = require("../nodes");
var Excerpt_1 = require("../parser/Excerpt");
/**
 * Implementation of DocNodeTransforms.trimSpacesInParagraphNodes()
 */
var TrimSpacesTransform = /** @class */ (function () {
    function TrimSpacesTransform() {
    }
    TrimSpacesTransform.transform = function (docParagraph) {
        var transformedNodes = [];
        // Whether the next nonempty node to be added needs a space before it
        var pendingSpace = false;
        // The DocPlainText node that we're currently accumulating
        var accumulatedPlainTextNode = undefined;
        // We always trim leading whitespace for a paragraph.  This flag gets set to true
        // as soon as nonempty content is encountered.
        var finishedSkippingLeadingSpaces = false;
        for (var _i = 0, _a = docParagraph.nodes; _i < _a.length; _i++) {
            var node = _a[_i];
            switch (node.kind) {
                case "PlainText" /* PlainText */:
                    var docPlainText = node;
                    var startedWithSpace = /^\s/.test(docPlainText.text);
                    var endedWithSpace = /\s$/.test(docPlainText.text);
                    var collapsedText = docPlainText.text.replace(/\s+/g, ' ').trim();
                    if (startedWithSpace && finishedSkippingLeadingSpaces) {
                        pendingSpace = true;
                    }
                    if (collapsedText.length > 0) {
                        if (accumulatedPlainTextNode) {
                            // If this node can't be merged, then eject it
                            if (!TrimSpacesTransform._canMergeExcerpts(accumulatedPlainTextNode.excerpt, docPlainText.excerpt)) {
                                transformedNodes.push(new nodes_1.DocPlainText(accumulatedPlainTextNode));
                                accumulatedPlainTextNode = undefined;
                            }
                        }
                        // If we haven't started an accumulatedPlainTextNode, create it now
                        if (!accumulatedPlainTextNode) {
                            accumulatedPlainTextNode = {
                                excerpt: undefined,
                                text: ''
                            };
                        }
                        if (pendingSpace) {
                            accumulatedPlainTextNode.text += ' ';
                            pendingSpace = false;
                        }
                        accumulatedPlainTextNode.text += collapsedText;
                        accumulatedPlainTextNode.excerpt = TrimSpacesTransform._mergeExcerpts(accumulatedPlainTextNode.excerpt, docPlainText.excerpt);
                        finishedSkippingLeadingSpaces = true;
                    }
                    if (endedWithSpace && finishedSkippingLeadingSpaces) {
                        pendingSpace = true;
                    }
                    break;
                case "SoftBreak" /* SoftBreak */:
                    if (finishedSkippingLeadingSpaces) {
                        pendingSpace = true;
                    }
                    break;
                default:
                    if (pendingSpace) {
                        // If we haven't started an accumulatedPlainTextNode, create it now
                        if (!accumulatedPlainTextNode) {
                            accumulatedPlainTextNode = {
                                excerpt: undefined,
                                text: ''
                            };
                        }
                        accumulatedPlainTextNode.text += ' ';
                        pendingSpace = false;
                    }
                    // Push the accumulated text
                    if (accumulatedPlainTextNode) {
                        transformedNodes.push(new nodes_1.DocPlainText(accumulatedPlainTextNode));
                        accumulatedPlainTextNode = undefined;
                    }
                    transformedNodes.push(node);
                    finishedSkippingLeadingSpaces = true;
            }
        }
        // Push the accumulated text
        if (accumulatedPlainTextNode) {
            transformedNodes.push(new nodes_1.DocPlainText(accumulatedPlainTextNode));
            accumulatedPlainTextNode = undefined;
        }
        var transformedParagraph = new nodes_1.DocParagraph({});
        transformedParagraph.appendNodes(transformedNodes);
        return transformedParagraph;
    };
    TrimSpacesTransform._canMergeExcerpts = function (currentExcerpt, followingExcerpt) {
        if (currentExcerpt === undefined || followingExcerpt === undefined) {
            return true;
        }
        if (!currentExcerpt.spacingAfterContent.isEmpty()
            || !followingExcerpt.spacingAfterContent.isEmpty()) {
            return false;
        }
        var currentSequence = currentExcerpt.content;
        var followingSequence = followingExcerpt.content;
        if (currentSequence.parserContext !== followingSequence.parserContext) {
            return false;
        }
        return currentSequence.endIndex === followingSequence.startIndex;
    };
    TrimSpacesTransform._mergeExcerpts = function (currentExcerpt, followingExcerpt) {
        if (currentExcerpt === undefined) {
            return followingExcerpt;
        }
        if (followingExcerpt === undefined) {
            return currentExcerpt;
        }
        if (!currentExcerpt.spacingAfterContent.isEmpty()
            || !followingExcerpt.spacingAfterContent.isEmpty()) {
            // This would be a program bug
            throw new Error('mergeExcerpts(): Cannot merge excerpts with spacingAfterContent');
        }
        var currentSequence = currentExcerpt.content;
        var followingSequence = followingExcerpt.content;
        if (currentSequence.parserContext !== followingSequence.parserContext) {
            // This would be a program bug
            throw new Error('mergeExcerpts(): Cannot merge excerpts with incompatible parser contexts');
        }
        if (currentSequence.endIndex !== followingSequence.startIndex) {
            // This would be a program bug
            throw new Error('mergeExcerpts(): Cannot merge excerpts that are not adjacent');
        }
        return new Excerpt_1.Excerpt({
            content: currentSequence.getNewSequence(currentSequence.startIndex, followingSequence.endIndex)
        });
    };
    return TrimSpacesTransform;
}());
exports.TrimSpacesTransform = TrimSpacesTransform;
//# sourceMappingURL=TrimSpacesTransform.js.map